from typing import Literal

import orjson


async def get_api_v5_copytrading_current_subpositions(self, **kwargs):
    """
    Retrieve lead positions that are not closed.
    Returns reverse chronological order with openTime
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-existing-lead-positions
    """
    endpoint = "/api/v5/copytrading/current-subpositions"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_subpositions_history(self, **kwargs):
    """
    Retrieve the completed lead position of the last 3 months.
    Returns reverse chronological order with subPosId
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-position-history
    """
    endpoint = "/api/v5/copytrading/subpositions-history"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_copytrading_algo_order(self, sub_pos_id: str, **kwargs):
    """
    Set TP/SL for the current lead position that are not closed.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-post-place-lead-stop-order
    """
    endpoint = "/api/v5/copytrading/algo-order"
    payload = {"subPosId": sub_pos_id, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_copytrading_close_subposition(self, sub_pos_id: str, **kwargs):
    """
    You can only close a lead position once a time.
    It is required to pass subPosId which can get from Get existing leading positions.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-post-close-lead-position
    """
    endpoint = "/api/v5/copytrading/close-subposition"
    payload = {"subPosId": sub_pos_id, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_instruments(self, **kwargs):
    """
    Retrieve instruments that are supported to lead by the platform. Retrieve instruments that the lead trader has set.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-leading-instruments
    """
    endpoint = "/api/v5/copytrading/instruments"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_copytrading_set_instruments(self, inst_id: str, **kwargs):
    """
    The leading trader can amend current leading instruments, need to set initial leading instruments while applying to become a leading trader.
    All non-leading instruments can't have position or pending orders for the current request when setting non-leading instruments as leading instruments.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-post-amend-leading-instruments
    """
    endpoint = "/api/v5/copytrading/set-instruments"
    payload = {"instId": inst_id, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_profit_sharing_details(self, **kwargs):
    """
    The leading trader gets profits shared details for the last 3 months.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-profit-sharing-details
    """
    endpoint = "/api/v5/copytrading/profit-sharing-details"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_total_profit_sharing(
    self, inst_type: str | None = None
):
    """
    The leading trader gets the total amount of profit shared since joining the platform.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-total-profit-sharing
    """
    endpoint = "/api/v5/copytrading/total-profit-sharing"
    payload = {}
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_unrealized_profit_sharing_details(
    self, inst_type: str | None = None
):
    """
    The leading trader gets the profit sharing details that are expected to be shared in the next settlement cycle.
    The unrealized profit sharing details will update once there copy position is closed.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-unrealized-profit-sharing-details
    """
    endpoint = "/api/v5/copytrading/unrealized-profit-sharing-details"
    payload = {}
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_total_unrealized_profit_sharing(
    self, inst_type: str | None = None
):
    """
    The leading trader gets the total unrealized amount of profit shared.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-total-unrealized-profit-sharing
    """
    endpoint = "/api/v5/copytrading/total-unrealized-profit-sharing"
    payload = {}
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_copytrading_apply_lead_trading(
    self,
    inst_id: str,
    inst_type: str | None = None,
):
    """
    Only ND broker sub-account whitelisted can apply for lead trader by this endpoint. It will be passed immediately.
    Please reach out to BD for help if you want to be whitelisted.
    For other accounts, e.g. ND main accounts and general main and sub-accounts, still need to apply on the web manually
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-post-apply-for-lead-trading
    """
    endpoint = "/api/v5/copytrading/apply-lead-trading"
    payload = {
        "instId": inst_id,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_copytrading_stop_lead_trading(self, inst_type: str | None = None):
    """
    It is used to stop lead trading for ND broker sub-account.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-post-stop-lead-trading
    """
    endpoint = "/api/v5/copytrading/stop-lead-trading"
    payload = {}
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_copytrading_amend_profit_sharing_ratio(
    self,
    profit_sharing_ratio: str,
    inst_id: str | None = None,
):
    """
    It is used to amend profit sharing ratio.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-post-amend-profit-sharing-ratio
    """
    endpoint = "/api/v5/copytrading/amend-profit-sharing-ratio"
    payload = {"profitSharingRatio": profit_sharing_ratio}
    if inst_id:
        payload["instId"] = inst_id
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_config(self):
    """
    Retrieve the configuration of the copy trading system.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-copy-trading-config
    """
    endpoint = "/api/v5/copytrading/config"
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def post_api_v5_copytrading_first_copy_settings(
    self,
    unique_code: str,
    copy_mgn_mode: Literal["cross", "isolated", "copy"],
    copy_inst_id_type: Literal["custom", "copy"],
    copy_total_amt: str,
    sub_pos_close_type: Literal["market_close", "copy_close", "manual_close"],
    **kwargs,
):
    """
    The first copy settings for the certain lead trader. You need to first copy settings after stopping copying.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-post-first-copy-settings
    """
    endpoint = "/api/v5/copytrading/first-copy-settings"
    payload = {
        "uniqueCode": unique_code,
        "copyMgnMode": copy_mgn_mode,
        "copyInstIdType": copy_inst_id_type,
        "copyTotalAmt": copy_total_amt,
        "subPosCloseType": sub_pos_close_type,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_copytrading_amend_copy_settings(
    self,
    unique_code: str,
    copy_mgn_mode: Literal["cross", "isolated", "copy"],
    copy_inst_id_type: Literal["custom", "copy"],
    copy_total_amt: str,
    sub_pos_close_type: Literal["market_close", "copy_close", "manual_close"],
    **kwargs,
):
    """
    You need to use this endpoint to amend copy settings
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-post-amend-copy-settings
    """
    endpoint = "/api/v5/copytrading/amend-copy-settings"
    payload = {
        "uniqueCode": unique_code,
        "copyMgnMode": copy_mgn_mode,
        "copyInstIdType": copy_inst_id_type,
        "copyTotalAmt": copy_total_amt,
        "subPosCloseType": sub_pos_close_type,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_copytrading_stop_copy_trading(
    self,
    unique_code: str,
    sub_pos_close_type: Literal["market_close", "copy_close", "manual_close"],
    inst_type: str | None = None,
):
    """
    You need to use this endpoint to stop copy trading
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-post-stop-copying
    """
    endpoint = "/api/v5/copytrading/stop-copy-trading"
    payload = {
        "uniqueCode": unique_code,
        "subPosCloseType": sub_pos_close_type,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_copy_settings(
    self,
    unique_code: str,
    inst_type: str | None = None,
):
    """
    Retrieve the copy settings about certain lead trader.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-copy-settings
    """
    endpoint = "/api/v5/copytrading/copy-settings"
    payload = {
        "uniqueCode": unique_code,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_current_lead_traders(
    self,
    inst_type: str | None = None,
):
    """
    Retrieve my lead traders.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-my-lead-traders
    """
    endpoint = "/api/v5/copytrading/current-lead-traders"
    payload = {}
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_public_config(self, inst_type: str | None = None):
    """
    Public endpoint. Retrieve copy trading parameter configuration information of copy settings
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-copy-trading-configuration
    """
    endpoint = "/api/v5/copytrading/public-config"
    payload = {}
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_copytrading_public_lead_traders(self, **kwargs):
    """
    Retrieve public lead traders.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-ranks
    """
    endpoint = "/api/v5/copytrading/public-lead-traders"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_copytrading_public_weekly_pnl(
    self,
    unique_code: str,
    inst_type: str | None = None,
):
    """
    Public endpoint. Retrieve lead trader weekly pnl. Results are returned in counter chronological order.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-weekly-pnl
    """
    endpoint = "/api/v5/copytrading/public-weekly-pnl"
    payload = {
        "uniqueCode": unique_code,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_copytrading_public_pnl(
    self,
    unique_code: str,
    last_days: Literal["1", "2", "3", "4"],
    inst_type: str | None = None,
):
    """
    Public endpoint. Retrieve lead trader daily pnl. Results are returned in counter chronological order.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-daily-pnl
    """
    endpoint = "/api/v5/copytrading/public-pnl"
    payload = {
        "uniqueCode": unique_code,
        "lastDays": last_days,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_copytrading_public_stats(
    self,
    unique_code: str,
    last_days: Literal["1", "2", "3", "4"],
    inst_type: str | None = None,
):
    """
    Public endpoint. Key data related to lead trader performance.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-stats
    """
    endpoint = "/api/v5/copytrading/public-stats"
    payload = {
        "uniqueCode": unique_code,
        "lastDays": last_days,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_copytrading_public_preference_currency(
    self,
    unique_code: str,
    inst_type: str | None = None,
):
    """
    Public endpoint. The most frequently traded crypto of this lead trader. Results are sorted by ratio from large to small.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-currency-preferences
    """
    endpoint = "/api/v5/copytrading/public-preference-currency"
    payload = {
        "uniqueCode": unique_code,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_copytrading_public_current_subpositions(
    self, unique_code: str, **kwargs
):
    """
    Public endpoint. Get current leading positions of lead trader
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-current-lead-positions
    """
    endpoint = "/api/v5/copytrading/public-current-subpositions"
    payload = {"uniqueCode": unique_code, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_copytrading_public_subpositions_history(
    self, unique_code: str, **kwargs
):
    """
    Public endpoint. Retrieve the lead trader completed leading position of the last 3 months.
    Returns reverse chronological order with subPosId.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-lead-position-history
    """
    endpoint = "/api/v5/copytrading/public-subpositions-history"
    payload = {"uniqueCode": unique_code, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_copytrading_public_copy_traders(self, unique_code: str, **kwargs):
    """
    Public endpoint. Retrieve copy trader coming from certain lead trader. Return according to pnl from high to low
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-copy-traders
    """
    endpoint = "/api/v5/copytrading/public-copy-traders"
    payload = {"uniqueCode": unique_code, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_copytrading_lead_traders(self, **kwargs):
    """
    Private endpoint. Retrieve lead trader ranks.
    For requests from the ND sub-account, under the same ND broker, this private endpoint can return ND lead trader information that the related public endpoint can't return.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-ranks-private
    """
    endpoint = "/api/v5/copytrading/lead-traders"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_weekly_pnl(
    self,
    unique_code: str,
    inst_type: str | None = None,
):
    """
    Private endpoint. Retrieve lead trader weekly pnl. Results are returned in counter chronological order.
    For requests from the ND sub-account, under the same ND broker, uniqueCode is supported for ND lead trader unique code by this endpoint, but the related public endpoint does not support it.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-weekly-pnl-private
    """
    endpoint = "/api/v5/copytrading/weekly-pnl"
    payload = {
        "uniqueCode": unique_code,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_pnl(
    self,
    unique_code: str,
    last_days: Literal["1", "2", "3", "4"],
    inst_type: str | None = None,
):
    """
    Private endpoint. Retrieve lead trader daily pnl. Results are returned in counter chronological order.
    For requests from the ND sub-account, under the same ND broker, uniqueCode is supported for ND lead trader unique code by this endpoint, but the related public endpoint does not support it.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-daily-pnl-private
    """
    endpoint = "/api/v5/copytrading/pnl"
    payload = {
        "uniqueCode": unique_code,
        "lastDays": last_days,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_stats(
    self,
    unique_code: str,
    last_days: Literal["1", "2", "3", "4"],
    inst_type: str | None = None,
):
    """
    Private endpoint. Key data related to lead trader performance.
    For requests from the ND sub-account, under the same ND broker, uniqueCode is supported for ND lead trader unique code by this endpoint, but the related public endpoint does not support it.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-stats-private
    """
    endpoint = "/api/v5/copytrading/stats"
    payload = {
        "uniqueCode": unique_code,
        "lastDays": last_days,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_preference_currency(
    self, unique_code: str, inst_type: str | None = None
):
    """
    Private endpoint. The most frequently traded crypto of this lead trader. Results are sorted by ratio from large to small.
    For requests from the ND sub-account, under the same ND broker, uniqueCode is supported for ND lead trader unique code by this endpoint, but the related public endpoint does not support it.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-lead-trader-currency-preferences-private
    """
    endpoint = "/api/v5/copytrading/preference-currency"
    payload = {
        "uniqueCode": unique_code,
    }
    if inst_type:
        payload["instType"] = inst_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_performance_current_subpositions(
    self, unique_code: str, **kwargs
):
    """
    Private endpoint. Get current leading positions of lead trader
    For requests from the ND sub-account, under the same ND broker, uniqueCode is supported for ND lead trader unique code by this endpoint, but the related public endpoint does not support it.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-performance-of-existing-lead-positions
    """
    endpoint = "/api/v5/copytrading/performance-current-subpositions"
    payload = {"uniqueCode": unique_code, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_performance_subpositions_history(
    self, unique_code: str, **kwargs
):
    """
    Private endpoint. Retrieve the lead trader completed leading position of the last 3 months.
    Returns reverse chronological order with subPosId.
    For requests from the ND sub-account, under the same ND broker, uniqueCode is supported for ND lead trader unique code by this endpoint, but the related public endpoint does not support it.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-performance-of-lead-position-history
    """
    endpoint = "/api/v5/copytrading/performance-subpositions-history"
    payload = {"uniqueCode": unique_code, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_copytrading_copy_traders(self, unique_code: str, **kwargs):
    """
    Private endpoint. Retrieve copy trader coming from certain lead trader. Return according to pnl from high to low
    For requests from the ND sub-account, under the same ND broker, uniqueCode is supported for ND lead trader unique code by this endpoint, but the related public endpoint does not support it.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-copy-trading-get-copy-traders
    """
    endpoint = "/api/v5/copytrading/copy-traders"
    payload = {"uniqueCode": unique_code, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)



from abc import ABC, abstractmethod

import aiohttp
import certifi
import orjson
import ssl
import asynciolimiter

import asyncio
from typing import Any
from typing import Callable, Literal, Optional

from aiolimiter import AsyncLimiter
from picows import (
    ws_connect,
    WSFrame,
    WSTransport,
    WSListener,
    WSMsgType,
    WSAutoPingStrategy,
)
from nexusbridge.utils import Log, LiveClock


class ApiClient(ABC):
    def __init__(
        self,
        api_key: str = None,
        secret: str = None,
        timeout: int = 10,
        max_rate: int | None = None,
    ):
        self._api_key = api_key
        self._secret = secret
        self._timeout = timeout
        self._log = Log.get_logger(type(self).__name__)
        self._ssl_context = ssl.create_default_context(cafile=certifi.where())
        self._session: Optional[aiohttp.ClientSession] = None
        self._clock = LiveClock()

        if max_rate:
            self._limiter = asynciolimiter.Limiter(rate=max_rate)
        else:
            self._limiter = None

    def _init_session(self):
        if self._session is None:
            timeout = aiohttp.ClientTimeout(total=self._timeout)
            tcp_connector = aiohttp.TCPConnector(
                ssl=self._ssl_context, enable_cleanup_closed=True
            )
            self._session = aiohttp.ClientSession(
                connector=tcp_connector, json_serialize=orjson.dumps, timeout=timeout
            )

    async def close_session(self):
        if self._session:
            await self._session.close()
            self._session = None


class Listener(WSListener):
    def __init__(
        self,
        logger,
        handler: Callable[..., Any],
        specific_ping_msg=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._log = logger
        self._specific_ping_msg = specific_ping_msg
        self._callback = handler
        self._clock = LiveClock()

    def send_user_specific_ping(self, transport: WSTransport):
        if self._specific_ping_msg:
            transport.send(WSMsgType.TEXT, self._specific_ping_msg)
            self._log.debug(f"Sent user specific ping {self._specific_ping_msg}")
        else:
            transport.send_ping()
            self._log.debug("Sent default ping.")

    def on_ws_connected(self, transport: WSTransport):
        self._log.debug("Connected to Websocket...")

    def on_ws_disconnected(self, transport: WSTransport):
        self._log.debug("Disconnected from Websocket.")

    def on_ws_frame(self, transport: WSTransport, frame: WSFrame):
        try:
            match frame.msg_type:
                case WSMsgType.PING:
                    # Only send pong if auto_pong is disabled
                    transport.send_pong(frame.get_payload_as_bytes())
                    return
                case WSMsgType.TEXT:
                    # Queue raw bytes for handler to decode
                    self._callback(frame.get_payload_as_bytes())
                    return
                case WSMsgType.CLOSE:
                    close_code = frame.get_close_code()
                    close_msg = frame.get_close_message()
                    self._log.warning(
                        f"Received close frame. Close code: {close_code}, Close message: {close_msg}"
                    )
                    return
        except Exception as e:
            self._log.error(f"Error processing message: {str(e)}")


class WSClient(ABC):
    def __init__(
        self,
        url: str,
        limiter: AsyncLimiter,
        handler: Callable[..., Any],
        specific_ping_msg: bytes = None,
        reconnect_interval: int = 1,
        ping_idle_timeout: int = 2,
        ping_reply_timeout: int = 1,
        auto_ping_strategy: Literal[
            "ping_when_idle", "ping_periodically"
        ] = "ping_when_idle",
        enable_auto_ping: bool = True,
        enable_auto_pong: bool = False,
    ):
        self._clock = LiveClock()
        self._url = url
        self._specific_ping_msg = specific_ping_msg
        self._reconnect_interval = reconnect_interval
        self._ping_idle_timeout = ping_idle_timeout
        self._ping_reply_timeout = ping_reply_timeout
        self._enable_auto_pong = enable_auto_pong
        self._enable_auto_ping = enable_auto_ping
        self._listener: Listener = None
        self._transport = None
        self._subscriptions = {}
        self._limiter = limiter
        self._handler = handler
        if auto_ping_strategy == "ping_when_idle":
            self._auto_ping_strategy = WSAutoPingStrategy.PING_WHEN_IDLE
        elif auto_ping_strategy == "ping_periodically":
            self._auto_ping_strategy = WSAutoPingStrategy.PING_PERIODICALLY
        self._log = Log.get_logger(type(self).__name__)

    @property
    def connected(self):
        return self._transport and self._listener

    async def _connect(self):
        if not self.connected:
            WSListenerFactory = lambda: Listener(
                self._log,
                handler=self._handler,
                specific_ping_msg=self._specific_ping_msg,
            )  # noqa: E731
            self._transport, self._listener = await ws_connect(
                WSListenerFactory,
                self._url,
                enable_auto_ping=self._enable_auto_ping,
                auto_ping_idle_timeout=self._ping_idle_timeout,
                auto_ping_reply_timeout=self._ping_reply_timeout,
                auto_ping_strategy=self._auto_ping_strategy,
                enable_auto_pong=self._enable_auto_pong,
            )

    async def connect(self):
        await self._connect()
        await self._connection_handler()

    async def _connection_handler(self):
        while True:
            try:
                if not self.connected:
                    await self._connect()
                    await self._resubscribe()
                await self._transport.wait_disconnected()
            except asyncio.CancelledError:
                self.disconnect()
                await asyncio.sleep(0.5)
                break
            except Exception as e:
                self._log.error(f"Connection error: {e}")
                self._log.debug("Websocket reconnecting...")
                self._transport, self._listener = None, None
                await asyncio.sleep(self._reconnect_interval)

    async def _send(self, payload: dict):
        await self._limiter.acquire()
        self._transport.send(WSMsgType.TEXT, orjson.dumps(payload))

    def disconnect(self):
        if self.connected:
            self._transport.disconnect()

    @abstractmethod
    async def _resubscribe(self):
        pass



import orjson
from typing import Any, Dict, Callable, Literal


async def post_v5_order_create(
    self,
    category: Literal["linear", "inverse", "spot", "option"],
    symbol: str,
    side: Literal["Buy", "Sell"],
    order_type: Literal["Market", "Limit"],
    qty: str,
    **kwargs,
):
    """
    This endpoint supports to create the order for Spot, Margin trading, USDT perpetual, USDC perpetual, USDC futures, Inverse Futures and Options.
    https://bybit-exchange.github.io/docs/v5/order/create-order
    """
    endpoint = "/v5/order/create"
    payload = {
        "category": category,
        "symbol": symbol,
        "side": side,
        "orderType": order_type,
        "qty": qty,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_order_amend(
    self,
    category: Literal["linear", "inverse", "spot", "option"],
    symbol: str,
    **kwargs,
):
    """
    Amend Order
    https://bybit-exchange.github.io/docs/v5/order/amend-order
    """
    endpoint = "/v5/order/amend"
    payload = {
        "category": category,
        "symbol": symbol,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_order_cancel(
    self,
    category: Literal["linear", "inverse", "spot", "option"],
    symbol: str,
    **kwargs,
):
    """
    Cancel Order
    https://bybit-exchange.github.io/docs/v5/order/cancel-order
    """
    endpoint = "/v5/order/cancel"
    payload = {
        "category": category,
        "symbol": symbol,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_order_realtime(
    self, category: Literal["linear", "inverse", "spot", "option"], **kwargs
):
    """
    Get Open & Closed Orders
    https://bybit-exchange.github.io/docs/v5/order/open-order
    """
    endpoint = "/v5/order/realtime"
    payload = {
        "category": category,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_order_cancel_all(
    self, category: Literal["linear", "inverse", "spot", "option"], **kwargs
):
    """
    Cancel All Orders
    https://bybit-exchange.github.io/docs/v5/order/cancel-all
    """
    endpoint = "/v5/order/cancel-all"
    payload = {
        "category": category,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_order_history(
    self, category: Literal["linear", "inverse", "spot", "option"], **kwargs
):
    """
    Get Order History
    https://bybit-exchange.github.io/docs/v5/order/order-list
    """
    endpoint = "/v5/order/history"
    payload = {
        "category": category,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_execution_list(
    self, category: Literal["linear", "inverse", "spot", "option"], **kwargs
):
    """
    Get Trade History
    https://bybit-exchange.github.io/docs/v5/order/execution
    """
    endpoint = "/v5/execution/list"
    payload = {
        "category": category,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_order_create_batch(
    self,
    category: Literal["linear", "inverse", "spot", "option"],
    request: list[Dict[str, Any]],
):
    """
    Batch Place Order
    https://bybit-exchange.github.io/docs/v5/order/batch-place
    """
    endpoint = "/v5/order/create-batch"
    payload = {
        "category": category,
        "request": request,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_order_amend_batch(
    self,
    category: Literal["linear", "inverse", "spot", "option"],
    request: list[Dict[str, Any]],
):
    """
    Batch Amend Order
    https://bybit-exchange.github.io/docs/v5/order/batch-amend
    """
    endpoint = "/v5/order/amend-batch"
    payload = {
        "category": category,
        "request": request,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_order_cancel_batch(
    self,
    category: Literal["linear", "inverse", "spot", "option"],
    request: list[Dict[str, Any]],
):
    """
    Batch Cancel Order
    https://bybit-exchange.github.io/docs/v5/order/batch-cancel
    """
    endpoint = "/v5/order/cancel-batch"
    payload = {
        "category": category,
        "request": request,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_order_spot_borrow_check(
    self,
    symbol: str,
    side: Literal["buy", "sell"],
    category: str = "spot",
):
    """
    Get Borrow Quota (Spot)
    https://bybit-exchange.github.io/docs/v5/order/spot-borrow-quota
    """
    endpoint = "/v5/order/spot-borrow-check"
    payload = {
        "symbol": symbol,
        "side": side,
        "category": category,
    }
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_order_disconnected_cancel_all(
    self,
    time_window: int,
    product: Literal["OPTIONS", "DERIVATIVES", "SPOT"] = "OPTIONS",
):
    """
    Set Disconnect Cancel All
    https://bybit-exchange.github.io/docs/v5/order/dcp
    """
    endpoint = "/v5/order/disconnected-cancel-all"
    payload = {
        "timeWindow": time_window,
        "product": product,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)



from typing import Literal


async def get_fapi_v3_balance(self, **kwargs):
    """
    Query account balance info
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Futures-Account-Balance-V3
    """
    endpoint = "/fapi/v3/balance"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v2_balance(self, **kwargs):
    """
    Query account balance info
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Futures-Account-Balance-V2
    """
    endpoint = "/fapi/v2/balance"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v3_account(self, **kwargs):
    """
    Get current account information. User in single-asset/ multi-assets mode will see different value, see comments in response section for detail.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Account-Information-V3
    """
    endpoint = "/fapi/v3/account"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v2_account(self, **kwargs):
    """
    Get current account information. User in single-asset/ multi-assets mode will see different value, see comments in response section for detail.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Account-Information-V2
    """
    endpoint = "/fapi/v2/account"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_commission_rate(self, symbol: str, **kwargs):
    """
    Get User Commission Rate
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/User-Commission-Rate
    """
    endpoint = "/fapi/v1/commissionRate"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_account_config(self, **kwargs):
    """
    Query account configuration
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Account-Config
    """
    endpoint = "/fapi/v1/accountConfig"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_symbol_config(self, **kwargs):
    """
    Get current account symbol configuration.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Symbol-Config
    """
    endpoint = "/fapi/v1/symbolConfig"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_rate_limit_order(self, **kwargs):
    """
    Query User Rate Limit
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Query-Rate-Limit
    """
    endpoint = "/fapi/v1/rateLimit/order"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_leverage_bracket(self, **kwargs):
    """
    Query user notional and leverage bracket on speicfic symbol
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Notional-and-Leverage-Brackets
    """
    endpoint = "/fapi/v1/leverageBracket"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_multi_assets_margin(self, **kwargs):
    """
    Get user's Multi-Assets mode (Multi-Assets Mode or Single-Asset Mode) on Every symbol
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-Current-Multi-Assets-Mode
    """
    endpoint = "/fapi/v1/multiAssetsMargin"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_position_side_dual(self, **kwargs):
    """
    Get user's position mode (Hedge Mode or One-way Mode ) on EVERY symbol
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-Current-Position-Mode
    """
    endpoint = "/fapi/v1/positionSide/dual"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_income(self, **kwargs):
    """
    Query income history
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-Income-History
    """
    endpoint = "/fapi/v1/income"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_api_trading_status(self, **kwargs):
    """
    Futures trading quantitative rules indicators, for more information on this, please refer to the Futures Trading Quantitative Rules
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Futures-Trading-Quantitative-Rules-Indicators
    """
    endpoint = "/fapi/v1/apiTradingStatus"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_income_asyn(self, start_time: int, end_time: int, **kwargs):
    """
    Get download id for futures transaction history
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-Download-Id-For-Futures-Transaction-History
    """
    endpoint = "/fapi/v1/income/asyn"
    payload = {"startTime": start_time, "endTime": end_time, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_income_asyn_id(self, download_id: str, **kwargs):
    """
    Get futures transaction history download link by Id
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-Futures-Transaction-History-Download-Link-by-Id
    """
    endpoint = "/fapi/v1/income/asyn/id"
    payload = {"downloadId": download_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_order_asyn(self, start_time: int, end_time: int, **kwargs):
    """
    Get Download Id For Futures Order History
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-Download-Id-For-Futures-Order-History
    """
    endpoint = "/fapi/v1/order/asyn"
    payload = {"startTime": start_time, "endTime": end_time, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_order_asyn_id(self, download_id: str, **kwargs):
    """
    Get futures order history download link by Id
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-Futures-Order-History-Download-Link-by-Id
    """
    endpoint = "/fapi/v1/order/asyn/id"
    payload = {"downloadId": download_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_trade_asyn(self, start_time: int, end_time: int, **kwargs):
    """
    Get download id for futures trade history
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-Download-Id-For-Futures-Trade-History
    """
    endpoint = "/fapi/v1/trade/asyn"
    payload = {"startTime": start_time, "endTime": end_time, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_trade_asyn_id(self, download_id: str, **kwargs):
    """
    Get futures trade download link by Id
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-Futures-Trade-Download-Link-by-Id
    """
    endpoint = "/fapi/v1/trade/asyn/id"
    payload = {"downloadId": download_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_fee_burn(self, fee_burn: Literal["true", "false"], **kwargs):
    """
    Change user's BNB Fee Discount (Fee Discount On or Fee Discount Off ) on EVERY symbol
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Toggle-BNB-Burn-On-Futures-Trade
    """
    endpoint = "/fapi/v1/feeBurn"
    payload = {"feeBurn": fee_burn, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)

    return raw


async def get_fapi_v1_fee_burn(self, **kwargs):
    """
    Get user's BNB Fee Discount (Fee Discount On or Fee Discount Off )
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/rest-api/Get-BNB-Burn-Status
    """
    endpoint = "/fapi/v1/feeBurn"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_pm_account_info(self, asset: str, **kwargs):
    """
    Get Classic Portfolio Margin current account information.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/portfolio-margin-endpoints
    """
    endpoint = "/fapi/v1/pmAccountInfo"
    payload = {"asset": asset, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw



async def get_api_v3_ping(self):
    """
    Test connectivity
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/general-endpoints
    """
    endpoint = "/api/v3/ping"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_api_v3_time(self):
    """
    Check server time
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/general-endpoints
    """
    endpoint = "/api/v3/time"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_api_v3_exchangeInfo(self, **kwargs):
    """
    Current exchange trading rules and symbol information
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/general-endpoints
    """
    endpoint = "/api/v3/exchangeInfo"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw



# Grid Trading
# Grid trading works by the simple strategy of buy low and sell high. After you set the parameters, the system automatically places orders at incrementally increasing or decreasing prices. Overall, the grid bot seeks to capitalize on normal price volatility by placing buy and sell orders at certain regular intervals above and below a predefined base price.
# The API endpoints of Grid Trading require authentication.
from typing import Literal

import orjson


async def post_api_v5_trading_bot_grid_order_algo(
    self,
    inst_id: str,
    algo_ord_type: Literal["grid", "contract_grid"],
    max_px: str,
    min_px: str,
    grid_num: str,
    trigger_param: list[dict[str, str]] | None = None,
    base_sz: str | None = None,
    quote_sz: str | None = None,
    **kwargs,
):
    """
    Create a grid strategy order
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-place-grid-algo-order
    """
    endpoint = "/api/v5/tradingBot/grid/order-algo"
    payload = {
        "instId": inst_id,
        "algoOrdType": algo_ord_type,
        "maxPx": max_px,
        "minPx": min_px,
        "gridNum": grid_num,
        **kwargs,
    }
    if base_sz:
        payload["baseSz"] = base_sz
    if quote_sz:
        payload["quoteSz"] = quote_sz
    if trigger_param:
        payload["triggerParams"] = trigger_param
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_amend_order_algo(
    self,
    inst_id: str,
    algo_id: str,
    trigger_param: list[dict[str, str]] | None = None,
    **kwargs,
):
    """
    Amend a grid strategy order
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-amend-grid-algo-order
    """
    endpoint = "/api/v5/tradingBot/grid/amend-order-algo"
    payload = {
        "instId": inst_id,
        "algoId": algo_id,
        **kwargs,
    }
    if trigger_param:
        payload["triggerParams"] = trigger_param
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_stop_order_algo(
    self,
    payload: list[dict[str, str]],
):
    """
    A maximum of 10 orders can be stopped per request.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-stop-grid-algo-order
    """
    endpoint = "/api/v5/tradingBot/grid/stop-order-algo"
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_close_position(
    self, algo_id: str, mkt_close: Literal[True, False], **kwargs
):
    """
    Close position when the contract grid stop type is 'keep position'.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-close-position-for-contract-grid
    """
    endpoint = "/api/v5/tradingBot/grid/close-position"
    payload = {
        "algoId": algo_id,
        "mktClose": mkt_close,
        "tag": "f50cdd72d3b6BCDE",
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_cancel_close_order(
    self, algo_id: str, ord_id: str
):
    """
    Cancel close position order for contract grid
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-cancel-close-position-order-for-contract-grid
    """
    endpoint = "/api/v5/tradingBot/grid/cancel-close-order"
    payload = {
        "algoId": algo_id,
        "ordId": ord_id,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_order_instant_trigger(
    self,
    algo_id: str,
):
    """
    Place an instant trigger order
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-instant-trigger-order
    """
    endpoint = "/api/v5/tradingBot/grid/order-instant-trigger"
    payload = {
        "algoId": algo_id,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_grid_orders_algo_pending(
    self, algo_ord_type: Literal["grid", "contract_grid"], **kwargs
):
    """
    Grid algo order list
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-get-grid-algo-order-list
    """
    endpoint = "/api/v5/tradingBot/grid/orders-algo-pending"
    payload = {
        "algoOrdType": algo_ord_type,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_grid_orders_algo_history(
    self, algo_ord_type: Literal["grid", "contract_grid"], **kwargs
):
    """
    Grid algo order history
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-get-grid-algo-order-history
    """
    endpoint = "/api/v5/tradingBot/grid/orders-algo-history"
    payload = {
        "algoOrdType": algo_ord_type,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_grid_orders_algo_details(
    self,
    algo_id: str,
    algo_ord_type: Literal["grid", "contract_grid"],
):
    """
    Grid algo order details
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-get-grid-algo-order-details
    """
    endpoint = "/api/v5/tradingBot/grid/orders-algo-details"
    payload = {
        "algoId": algo_id,
        "algoOrdType": algo_ord_type,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_grid_sub_orders(
    self,
    algo_id: str,
    algo_ord_type: Literal["grid", "contract_grid"],
    _type: Literal["live", "filled"],
    **kwargs,
):
    """
    Grid algo sub orders
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-get-grid-algo-sub-orders
    """
    endpoint = "/api/v5/tradingBot/grid/sub-orders"
    payload = {
        "algoId": algo_id,
        "algoOrdType": algo_ord_type,
        "type": _type,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_grid_positions(
    self,
    algo_id: str,
    algo_ord_type: Literal["grid", "contract_grid"],
):
    """
    Grid algo order positions
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-get-grid-algo-order-positions
    """
    endpoint = "/api/v5/tradingBot/grid/positions"
    payload = {
        "algoId": algo_id,
        "algoOrdType": algo_ord_type,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_withdraw_income(
    self,
    algo_id: str,
):
    """
    Spot grid withdraw income
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-spot-grid-withdraw-income
    """
    endpoint = "/api/v5/tradingBot/grid/withdraw-income"
    payload = {
        "algoId": algo_id,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_compute_margin_balance(
    self,
    algo_id: str,
    _type: Literal["add", "reduce"],
    amt: str | None = None,
):
    """
    Compute margin balance
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-compute-margin-balance
    """
    endpoint = "/api/v5/tradingBot/grid/compute-margin-balance"
    payload = {
        "algoId": algo_id,
        "type": _type,
    }
    if amt:
        payload["amt"] = amt

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_margin_balance(
    self,
    algo_id: str,
    _type: Literal["add", "reduce"],
    amt: str | None = None,
    percent: str | None = None,
):
    """
    Adjust margin balance
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-adjust-margin-balance
    """
    endpoint = "/api/v5/tradingBot/grid/margin-balance"
    payload = {
        "algoId": algo_id,
        "type": _type,
    }
    if amt:
        payload["amt"] = amt
    if percent:
        payload["percent"] = percent
    if not amt and not percent:
        raise ValueError("amt or percent must be provided")
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_adjust_investment(
    self,
    algo_id: str,
    amt: str,
):
    """
    It is used to add investment and only applicable to contract gird.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-add-investment
    """
    endpoint = "/api/v5/tradingBot/grid/adjust-investment"
    payload = {
        "algoId": algo_id,
        "amt": amt,
    }

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_grid_ai_param(
    self, algo_ord_type: Literal["grid", "contract_grid"], inst_id: str, **kwargs
):
    """
    Get grid AI parameters
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-get-grid-ai-parameter-public
    """
    endpoint = "/api/v5/tradingBot/grid/ai-param"
    payload = {
        "algoOrdType": algo_ord_type,
        "instId": inst_id,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_grid_min_investment(
    self,
    inst_id: str,
    algo_ord_type: Literal["grid", "contract_grid"],
    max_px: str,
    min_px: str,
    grid_num: str,
    run_type: Literal["1", "2"],
    investment_data: list[dict[str, str]] | None = None,
    **kwargs,
):
    """
    Compute the minimum investment
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-post-compute-min-investment-public
    """
    endpoint = "/api/v5/tradingBot/grid/min-investment"
    payload = {
        "instId": inst_id,
        "algoOrdType": algo_ord_type,
        "maxPx": max_px,
        "minPx": min_px,
        "gridNum": grid_num,
        "runType": run_type,
        **kwargs,
    }
    if investment_data:
        payload["investmentData"] = investment_data
    raw = await self._fetch("POST", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_rsi_back_testing(
    self, inst_id: str, timeframe: str, thold: str, time_period: str, **kwargs
):
    """
    RSI back testing (public)
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-get-rsi-back-testing-public
    """
    endpoint = "/api/v5/tradingBot/public/rsi-back-testing"
    payload = {
        "instId": inst_id,
        "timeframe": timeframe,
        "thold": thold,
        "timePeriod": time_period,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_grid_quantity(
    self,
    inst_id: str,
    run_type: Literal["1", "2"],
    algo_ord_type: Literal["grid", "contract_grid"],
    max_px: str,
    min_px: str,
    lever: str | None = None,
):
    """
    Get grid quantity
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-grid-trading-get-max-grid-quantity-public
    """
    endpoint = "/api/v5/tradingBot/grid/grid-quantity"
    payload = {
        "instId": inst_id,
        "runType": run_type,
        "algoOrdType": algo_ord_type,
        "maxPx": max_px,
        "minPx": min_px,
    }

    if algo_ord_type == "contract_grid":
        if not lever:
            raise ValueError("lever must be provided")
        payload["lever"] = lever
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)



from typing import Literal

from nexusbridge.binance.constants import Interval, ContractType


async def get_fapi_v1_ping(
    self,
):
    """
    Test connectivity
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api
    """
    endpoint = "/fapi/v1/ping"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_fapi_v1_time(self):
    """
    Check server time
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Check-Server-Time
    """
    endpoint = "/fapi/v1/time"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_fapi_v1_exchange_info(
    self,
):
    """
    Current exchange trading rules and symbol information
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Exchange-Information
    """
    endpoint = "/fapi/v1/exchangeInfo"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_fapi_v1_depth(
    self,
    symbol: str,
    limit: int = 500,
):
    """
    Query symbol orderbook
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Order-Book
    """
    endpoint = "/fapi/v1/depth"
    payload = {"symbol": symbol, "limit": limit}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_trades(
    self,
    symbol: str,
    limit: int = 500,
):
    """
    Get recent market trades
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Recent-Trades-List
    """
    endpoint = "/fapi/v1/trades"
    payload = {"symbol": symbol, "limit": limit}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_historical_trades(self, symbol: str, **kwargs):
    """
    Get older market historical trades.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Old-Trades-Lookup
    """
    endpoint = "/fapi/v1/historicalTrades"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_agg_trades(self, symbol: str, **kwargs):
    """
    Get compressed, aggregate market trades. Market trades that fill in 100ms with the same price and the same taking side will have the quantity aggregated.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Compressed-Aggregate-Trades-List
    """
    endpoint = "/fapi/v1/aggTrades"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_klines(self, symbol: str, interval: Interval, **kwargs):
    """
    Kline/candlestick bars for a symbol. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Kline-Candlestick-Data
    """
    endpoint = "/fapi/v1/klines"
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_continuous_klines(
    self, pair: str, contract_type: ContractType, interval: Interval, **kwargs
):
    """
    Kline/candlestick bars for a specific contract type of a pair. Klines are uniquely identified by their open time.
    https://binance-docs.github.io/apidocs/futures/en/#continuous-contract-kline-candlestick-data
    """
    endpoint = "/fapi/v1/continuousKlines"
    payload = {
        "pair": pair,
        "contractType": contract_type.value,
        "interval": interval.value,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_index_price_klines(self, pair: str, interval: Interval, **kwargs):
    """
    Kline/candlestick bars for the index price of a pair. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Index-Price-Kline-Candlestick-Data
    """
    endpoint = "/fapi/v1/indexPriceKlines"
    payload = {"pair": pair, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_mark_price_klines(
    self, symbol: str, interval: Interval, **kwargs
):
    """
    Kline/candlestick bars for the mark price of a symbol. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Mark-Price-Kline-Candlestick-Data
    """
    endpoint = "/fapi/v1/markPriceKlines"
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_premium_index_klines(
    self, symbol: str, interval: Interval, **kwargs
):
    """
    Premium index kline bars of a symbol. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Premium-Index-Kline-Data
    """
    endpoint = "/fapi/v1/premiumIndexKlines"
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_premium_index(
    self,
    symbol: str,
):
    """
    Mark Price and Funding Rate
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Mark-Price
    """
    endpoint = "/fapi/v1/premiumIndex"
    payload = {
        "symbol": symbol,
    }
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_funding_rate(self, **kwargs):
    """
    Get Funding Rate History
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Get-Funding-Rate-History
    """
    endpoint = "/fapi/v1/fundingRate"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_funding_info(self):
    """
    Query funding rate info for symbols that had FundingRateCap/ FundingRateFloor / fundingIntervalHours adjustment
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Get-Funding-Rate-Info
    """
    endpoint = "/fapi/v1/fundingInfo"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_fapi_v1_ticker_24hr(self, symbol: str = None):
    """
    24 hour rolling window price change statistics.
    Careful when accessing this with no symbol.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/24hr-Ticker-Price-Change-Statistics
    """
    endpoint = "/fapi/v1/ticker/24hr"
    payload = {}
    if symbol:
        payload["symbol"] = symbol
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_ticker_price(self, symbol: str = None):
    """
    Latest price for a symbol or symbols.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/latestPrice
    """
    endpoint = "/fapi/v1/ticker/price"
    payload = {}
    if symbol:
        payload["symbol"] = symbol
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v2_ticker_price(self, symbol: str = None):
    """
    Latest price for a symbol or symbols.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Symbol-Price-Ticker-v2
    """
    endpoint = "/fapi/v2/ticker/price"
    payload = {}
    if symbol:
        payload["symbol"] = symbol
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_book_ticker(self, symbol: str = None):
    """
    Best price/qty on the order book for a symbol or symbols.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Symbol-Order-Book-Ticker
    """
    endpoint = "/fapi/v1/ticker/bookTicker"
    payload = {}
    if symbol:
        payload["symbol"] = symbol
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_delivery_price(self, pair: str):
    """
    Latest price for a symbol or symbols.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Delivery-Price
    """
    endpoint = "/fapi/v1/deliveryPrice"
    payload = {"pair": pair}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_open_interest(self, symbol: str):
    """
    Get present open interest of a specific symbol.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Open-Interest
    """
    endpoint = "/fapi/v1/openInterest"
    payload = {"symbol": symbol}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_open_interest_hist(
    self,
    symbol: str,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    Open Interest Statistics
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Open-Interest-Statistics
    """
    endpoint = "/fapi/v1/openInterestHist"
    payload = {"symbol": symbol, "period": period, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_futures_data_top_long_short_position_ratio(
    self,
    symbol: str,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    The proportion of net long and net short positions to total open positions of the top 20% users with the highest margin balance. Long Position % = Long positions of top traders / Total open positions of top traders Short Position % = Short positions of top traders / Total open positions of top traders Long/Short Ratio (Positions) = Long Position % / Short Position %
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Top-Trader-Long-Short-Ratio
    """
    endpoint = "/futures/data/topLongShortPositionRatio"
    payload = {"symbol": symbol, "period": period, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_futures_data_top_long_short_account_ratio(
    self,
    symbol: str,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    The proportion of net long and net short accounts to total accounts of the top 20% users with the highest margin balance. Each account is counted once only. Long Account % = Accounts of top traders with net long positions / Total accounts of top traders with open positions Short Account % = Accounts of top traders with net short positions / Total accounts of top traders with open positions Long/Short Ratio (Accounts) = Long Account % / Short Account %
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Top-Long-Short-Account-Ratio
    """
    endpoint = "/futures/data/topLongShortAccountRatio"
    payload = {"symbol": symbol, "period": period, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_futures_data_global_long_short_account_ratio(
    self,
    symbol: str,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    Query symbol Long/Short Ratio
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Long-Short-Ratio
    """
    endpoint = "/futures/data/globalLongShortAccountRatio"
    payload = {"symbol": symbol, "period": period, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_futures_data_taker_long_short_ratio(
    self,
    symbol: str,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    Taker Buy/Sell Volume
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Taker-BuySell-Volume
    """
    endpoint = "/futures/data/takerlongshortRatio"
    payload = {"symbol": symbol, "period": period, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_lvt_klines(self, symbol: str, interval: Interval, **kwargs):
    """
    Kline/candlestick bars for a symbol. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Kline-Candlestick-Data
    """
    endpoint = "/fapi/v1/lvtKlines"
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw


async def get_fapi_v1_index_info(self, symbol: str | None = None):
    """
    Query composite index symbol information
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Composite-Index-Symbol-Information
    """
    endpoint = "/fapi/v1/indexInfo"
    payload = {}
    if symbol:
        payload["symbol"] = symbol
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_fapi_v1_asset_index(self, symbol: str | None = None):
    """
    asset index for Multi-Assets mode
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Multi-Assets-Mode-Asset-Index
    """
    endpoint = "/fapi/v1/assetIndex"
    payload = {}
    if symbol:
        payload["symbol"] = symbol
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_fapi_v1_constituents(self, symbol: str):
    """
    Query index price constituents
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/rest-api/Index-Constituents
    """
    endpoint = "/fapi/v1/constituents"
    payload = {"symbol": symbol}
    raw = await self._fetch("GET", endpoint, signed=False, payload=payload)
    return raw



from typing import Literal

import orjson


async def get_api_v5_public_instruments(
    self, inst_type: Literal["SPOT", "MARGIN", "FUTURES", "SWAP", "OPTION"], **kwargs
):
    """
    Retrieve a list of instruments with open contracts.
    https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-instruments
    """
    endpoint = "/api/v5/public/instruments"
    payload = {
        "instType": inst_type,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_delivery_exercise_history(
    self, inst_type: Literal["FUTURES", "OPTION"], **kwargs
):
    """
    Retrieve delivery records of Futures and exercise records of Options in the last 3 months.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-delivery-exercise-history
    """
    endpoint = "/api/v5/public/delivery-exercise-history"
    payload = {
        "instType": inst_type,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_open_interest(
    self, inst_type: Literal["FUTURES", "SWAP", "OPTION"], **kwargs
):
    """
    Retrieve the total open interest for contracts on OKX.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-open-interest
    """
    endpoint = "/api/v5/public/open-interest"
    payload = {
        "instType": inst_type,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_funding_rate(
    self,
    inst_id: str,
):
    """
    Retrieve funding rate.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-funding-rate
    """
    endpoint = "/api/v5/public/funding-rate"
    payload = {
        "instId": inst_id,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_funding_rate_history(self, inst_id: str, **kwargs):
    """
    Retrieve funding rate history. This endpoint can retrieve data from the last 3 months.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-funding-rate-history
    """
    endpoint = "/api/v5/public/funding-rate-history"
    payload = {
        "instId": inst_id,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_price_limit(
    self,
    inst_id: str,
):
    """
    Retrieve the highest buy limit and lowest sell limit of the instrument.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-limit-price
    """
    endpoint = "/api/v5/public/price-limit"
    payload = {
        "instId": inst_id,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_opt_summary(self, **kwargs):
    """
    Retrieve option market data.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-option-market-data
    """
    endpoint = "/api/v5/public/opt-summary"
    payload = {
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_estimated_price(self, inst_id: str):
    """
    Retrieve the estimated delivery price which will only have a return value one hour before the delivery/exercise.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-estimated-delivery-exercise-price
    """
    endpoint = "/api/v5/public/estimated-price"
    payload = {
        "instId": inst_id,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_discount_rate_interest_free_quota(self, **kwargs):
    """
    Retrieve discount rate level and interest-free quota.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-discount-rate-and-interest-free-quota
    """
    endpoint = "/api/v5/public/discount-rate-interest-free-quota"
    payload = {
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_time(self):
    """
    Retrieve API server time.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-system-time
    """
    endpoint = "/api/v5/public/time"
    raw = await self._fetch("GET", endpoint, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_mark_price(
    self, inst_id: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"], **kwargs
):
    """
    Retrieve mark price.
    We set the mark price based on the SPOT index and at a reasonable basis to prevent individual users from manipulating the market and causing the contract price to fluctuate.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-mark-price
    """
    endpoint = "/api/v5/public/mark-price"
    payload = {
        "instId": inst_id,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_position_tiers(
    self,
    inst_type: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"],
    td_mode: Literal["cross", "isolated"],
    **kwargs,
):
    """
    Retrieve position tiers information, maximum leverage depends on your borrowings and margin ratio.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-position-tiers
    """

    endpoint = "/api/v5/public/position-tiers"
    payload = {
        "instType": inst_type,
        "tdMode": td_mode,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_interest_rate_loan_quota(self):
    """
    Retrieve interest rate
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-interest-rate-and-loan-quota
    """
    endpoint = "/api/v5/public/interest-rate-loan-quota"

    raw = await self._fetch("GET", endpoint, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_underlying(
    self, inst_type: Literal["SWAP", "FUTURES", "OPTION"]
):
    """
    Retrieve underlying index information.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-underlying-index
    """
    endpoint = "/api/v5/public/underlying"
    payload = {
        "instType": inst_type,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_insurance_fund(
    self, inst_type: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"], **kwargs
):
    """
    Get insurance fund balance information
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-insurance-fund
    """
    endpoint = "/api/v5/public/insurance-fund"
    payload = {
        "instType": inst_type,
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_convert_contract_coin(
    self, inst_id: str, sz: str, **kwargs
):
    """
    Convert the crypto value to the number of contracts, or vice versa
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-insurance-fund
    """
    endpoint = "/api/v5/public/convert-contract-coin"
    payload = {
        "instId": inst_id,
        "sz": sz,
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_instrument_tick_bands(
    self,
    inst_type: str = "OPTION",
    inst_family: str | None = None,
):
    """
    Get option tick bands information
    https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-option-tick-bands
    """
    endpoint = "/api/v5/public/instrument-tick-bands"
    payload = {
        "instType": inst_type,
    }
    if inst_family:
        payload["instFamily"] = inst_family
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_premium_history(self, inst_id: str, **kwargs):
    """
    It will return premium data in the past 6 months.
    https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-premium-history
    """
    endpoint = "/api/v5/public/premium-history"
    payload = {
        "instId": inst_id,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_index_tickers(self, **kwargs):
    """
    Retrieve index tickers.
    https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-index-tickers
    """
    endpoint = "/api/v5/market/index-tickers"
    payload = {
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_index_candles(self, inst_id: str, **kwargs):
    """
    Retrieve the candlestick charts of the index. This endpoint can retrieve the latest 1,440 data entries. Charts are returned in groups based on the requested bar.
    https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-index-candlesticks
    """
    endpoint = "/api/v5/market/index-candles"
    payload = {
        "instId": inst_id,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_history_index_candles(self, inst_id: str, **kwargs):
    """
    Retrieve the candlestick charts of the index from recent years.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-index-candlesticks-history
    """
    endpoint = "/api/v5/market/history-index-candles"
    payload = {
        "instId": inst_id,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_mark_price_candles(self, inst_id: str, **kwargs):
    """
    Retrieve the candlestick charts of mark price. This endpoint can retrieve the latest 1,440 data entries. Charts are returned in groups based on the requested bar.
    https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-mark-price-candlesticks
    """
    endpoint = "/api/v5/market/mark-price-candles"
    payload = {
        "instId": inst_id,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_history_mark_price_candles(self, inst_id: str, **kwargs):
    """
    Retrieve the candlestick charts of mark price from recent years.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-mark-price-candlesticks-history
    """
    endpoint = "/api/v5/market/history-mark-price-candles"
    payload = {
        "instId": inst_id,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_open_oracle(self):
    """
    Get the crypto price of signing using Open Oracle smart contract.
    https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-oracle
    """
    endpoint = "/api/v5/market/open-oracle"

    raw = await self._fetch("GET", endpoint, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_exchange_rate(self):
    """
    This interface provides the average exchange rate data for 2 weeks
    https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-exchange-rate
    """
    endpoint = "/api/v5/market/exchange-rate"

    raw = await self._fetch("GET", endpoint, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_index_components(self, index: str):
    """
    Get the index component information data on the market
    https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-index-components
    """
    endpoint = "/api/v5/market/index-components"
    payload = {
        "index": index,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_economic_calendar(self, **kwargs):
    """
    Get the macro-economic calendar data within 3 months. Historical data from 3 months ago is only available to users with trading fee tier VIP1 and above.
    https://www.okx.com/docs-v5/en/?python#public-data-rest-api-get-economic-calendar-data
    """
    endpoint = "/api/v5/public/economic-calendar"
    payload = {
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)



import json

from nexusbridge.binance.constants import Interval


async def get_api_v3_depth(self, symbol: str, limit: int = 100):
    """
    Order book
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/depth"
    payload = {"symbol": symbol, "limit": limit}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_trades(self, symbol: str, limit: int = 500):
    """
    Recent trades list
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/trades"
    payload = {"symbol": symbol, "limit": limit}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_historical_trades(self, symbol: str, **kwargs):
    """
    Old trade lookup
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/historicalTrades"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_klines(self, symbol: str, interval: Interval, **kwargs):
    """
    Kline/candlestick bars for a symbol. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/klines"
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_ui_klines(self, symbol: str, interval: Interval, **kwargs):
    """
    The request is similar to klines having the same parameters and response.
    uiKlines return modified kline data, optimized for presentation of candlestick charts.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/uiKlines"
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_avgPrice(self, symbol: str):
    """
    Current average price for a symbol.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/avgPrice"
    payload = {"symbol": symbol}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_ticker_24hr(self, **kwargs):
    """
    24 hour rolling window price change statistics. Careful when accessing this with no symbol.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/ticker/24hr"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_ticker_tradingDay(self, symbols: str | list[str], **kwargs):
    """
    Price change statistics for a trading day.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/ticker/tradingDay"
    payload = {**kwargs}
    if isinstance(symbols, str):
        payload["symbol"] = symbols
    else:
        payload["symbols"] = str(json.dumps(symbols)).replace(" ", "")
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_ticker_price(self, **kwargs):
    """
    Latest price for a symbol or symbols.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/ticker/price"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_ticker_bookTicker(self, **kwargs):
    """
    Best price/qty on the order book for a symbol or symbols.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/ticker/bookTicker"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_api_v3_ticker(self, symbols: str | list[str], **kwargs):
    """
    Rolling window price change statistics
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/market-data-endpoints
    """
    endpoint = "/api/v3/ticker"
    payload = {**kwargs}
    if isinstance(symbols, str):
        payload["symbol"] = symbols
    else:
        payload["symbols"] = str(json.dumps(symbols)).replace(" ", "")
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw



import orjson
from typing import Any, Dict, Literal


async def get_v5_position_list(
    self, category: Literal["linear", "inverse", "option"], **kwargs
):
    """
    Get Position Info
    https://bybit-exchange.github.io/docs/v5/position
    """
    endpoint = "/v5/position/list"
    payload = {
        "category": category,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_set_leverage(
    self,
    category: Literal["linear", "inverse"],
    symbol: str,
    buy_leverage: str,
    sell_leverage: str,
):
    """
    Set Leverage
    https://bybit-exchange.github.io/docs/v5/position/leverage
    """
    endpoint = "/v5/position/set-leverage"
    payload = {
        "category": category,
        "symbol": symbol,
        "buyLeverage": buy_leverage,
        "sellLeverage": sell_leverage,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_switch_isolated(
    self,
    symbol: str,
    category: Literal["inverse", "linear"],
    trade_mode: Literal[0, 1],
    buy_leverage: str,
    sell_leverage: str,
):
    """
    Switch Cross/Isolated Margin
    https://bybit-exchange.github.io/docs/v5/position/cross-isolate
    """
    endpoint = "/v5/position/switch-isolated"
    payload = {
        "symbol": symbol,
        "category": category,
        "tradeMode": trade_mode,
        "buyLeverage": buy_leverage,
        "sellLeverage": sell_leverage,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_switch_mode(
    self, category: Literal["inverse", "linear"], mode: Literal[0, 3], **kwargs
):
    """
    Switch Position Mode
    https://bybit-exchange.github.io/docs/v5/position/position-mode
    """
    endpoint = "/v5/position/switch-mode"
    payload = {
        "category": category,
        "mode": mode,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_trading_stop(
    self,
    category: Literal["linear", "inverse"],
    symbol: str,
    position_idx: Literal[0, 1, 2],
    **kwargs,
):
    """
    Set Trading-Stop
    https://bybit-exchange.github.io/docs/v5/position/trading-stop
    """
    endpoint = "/v5/position/trading-stop"
    payload = {
        "category": category,
        "symbol": symbol,
        "positionIdx": position_idx,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_set_auto_add_margin(
    self,
    symbol: str,
    auto_add_margin: Literal[0, 1],
    position_idx: Literal[0, 1, 2] | None = None,
    category: str = "linear",
):
    """
    Set Auto Add Margin
    https://bybit-exchange.github.io/docs/v5/position/auto-add-margin
    """
    endpoint = "/v5/position/set-auto-add-margin"
    payload = {
        "symbol": symbol,
        "autoAddMargin": auto_add_margin,
        "category": category,
    }
    if position_idx:
        payload["positionIdx"] = position_idx
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_add_margin(
    self,
    category: Literal["linear", "inverse"],
    symbol: str,
    margin: str,
    position_idx: Literal[0, 1, 2] | None = None,
):
    """
    Add Or Reduce Margin
    https://bybit-exchange.github.io/docs/v5/position/manual-add-margin
    """
    endpoint = "/v5/position/add-margin"
    payload = {
        "symbol": symbol,
        "margin": margin,
        "category": category,
    }
    if position_idx:
        payload["positionIdx"] = position_idx
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_position_closed_pnl(
    self, category: Literal["linear", "inverse"], **kwargs
):
    """
    Get Closed PnL
    https://bybit-exchange.github.io/docs/v5/position/close-pnl
    """
    endpoint = "/v5/position/closed-pnl"
    payload = {
        "category": category,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_move_positions(
    self, from_uid: str, to_uid: str, _list: list[dict[str, Any]]
):
    """
    Move Position
    https://bybit-exchange.github.io/docs/v5/position/move-position
    """
    endpoint = "/v5/position/move-positions"
    payload = {
        "fromUid": from_uid,
        "toUid": to_uid,
        "list": _list,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_position_move_history(self, **kwargs):
    """
    Get Move Position History
    https://bybit-exchange.github.io/docs/v5/position/move-position-history
    """
    endpoint = "/v5/position/move-history"
    payload = {
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_confirm_pending_mmr(
    self,
    symbol: str,
    category: Literal["linear", "inverse"],
):
    """
    Confirm New Risk Limit
    https://bybit-exchange.github.io/docs/v5/position/confirm-mmr
    """
    endpoint = "/v5/position/confirm-pending-mmr"
    payload = {
        "symbol": symbol,
        "category": category,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_set_tpsl_mode(
    self,
    symbol: str,
    category: Literal["linear", "inverse"],
    tp_sl_mode: Literal["Full", "Partial"],
):
    """
    Set TP/SL Mode
    https://bybit-exchange.github.io/docs/v5/position/tpsl-mode
    """
    endpoint = "/v5/position/set-tpsl-mode"
    payload = {
        "symbol": symbol,
        "category": category,
        "tpSlMode": tp_sl_mode,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_position_set_risk_limit(
    self,
    symbol: str,
    category: Literal["linear", "inverse"],
    risk_id: int,
    position_idx: Literal[0, 1, 2] | None = None,
):
    """
    Set Risk Limit
    https://bybit-exchange.github.io/docs/v5/position/set-risk-limit
    """
    endpoint = "/v5/position/set-risk-limit"
    payload = {
        "symbol": symbol,
        "category": category,
        "riskId": risk_id,
    }
    if position_idx:
        payload["positionIdx"] = position_idx
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)



from enum import Enum, unique


class OkxUrl(Enum):
    LIVE = 0
    AWS = 1
    DEMO = 2

    @property
    def base_url(self):
        return REST_URLS[self]

    @property
    def is_testnet(self):
        return self == OkxUrl.DEMO


REST_URLS = {
    OkxUrl.LIVE: "https://www.okx.com",
    OkxUrl.AWS: "https://aws.okx.com",
    OkxUrl.DEMO: "https://www.okx.com",
}


class OkxInstrumentType(Enum):
    SPOT = "SPOT"
    MARGIN = "MARGIN"
    SWAP = "SWAP"
    FUTURES = "FUTURES"
    OPTION = "OPTION"
    ANY = "ANY"


class OkxInstrumentFamily(Enum):
    FUTURES = "FUTURES"
    SWAP = "SWAP"
    OPTION = "OPTION"


@unique
class OkxTdMode(Enum):
    CASH = "cash"  # 现货
    CROSS = "cross"  # 全仓
    ISOLATED = "isolated"  # 逐仓
    SPOT_ISOLATED = "spot_isolated"  # 现货逐仓


@unique
class OkxPositionSide(Enum):
    LONG = "long"
    SHORT = "short"
    NET = "net"
    NONE = ""


@unique
class OkxOrderSide(Enum):
    BUY = "buy"
    SELL = "sell"


class OkxTimeInForce(Enum):
    IOC = "ioc"
    GTC = "gtc"
    FOK = "fok"


@unique
class OkxOrderType(Enum):
    MARKET = "market"
    LIMIT = "limit"
    POST_ONLY = "post_only"  # limit only, requires "px" to be provided
    FOK = "fok"  # market order if "px" is not provided, otherwise limit order
    IOC = "ioc"  # market order if "px" is not provided, otherwise limit order
    OPTIMAL_LIMIT_IOC = (
        "optimal_limit_ioc"  # Market order with immediate-or-cancel order
    )
    MMP = "mmp"  # Market Maker Protection (only applicable to Option in Portfolio Margin mode)
    MMP_AND_POST_ONLY = "mmp_and_post_only"  # Market Maker Protection and Post-only order(only applicable to Option in Portfolio Margin mode)


@unique
class OkxOrderStatus(Enum):  # "state"
    CANCELED = "canceled"
    LIVE = "live"
    PARTIALLY_FILLED = "partially_filled"
    FILLED = "filled"
    MMP_CANCELED = "mmp_canceled"


class OkxAccountType(Enum):
    LIVE = 0
    AWS = 1
    DEMO = 2

    @property
    def exchange_id(self):
        return "okx"

    @property
    def is_testnet(self):
        return self == OkxAccountType.DEMO

    @property
    def stream_url(self):
        return STREAM_URLS[self]


class ChannelKind(Enum):
    PUBLIC = "public"
    PRIVATE = "private"
    BUSINESS = "business"


STREAM_URLS = {
    OkxAccountType.LIVE: "wss://ws.okx.com:8443/ws",
    OkxAccountType.AWS: "wss://wsaws.okx.com:8443/ws",
    OkxAccountType.DEMO: "wss://wspap.okx.com:8443/ws",
}



from typing import Literal


async def agg_trade(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    The Aggregate Trade Streams push market trade information that is aggregated for fills with same price and taking side every 100 milliseconds.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Aggregate-Trade-Streams
    """
    stream_name = "{}@aggTrade".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def index_price(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    Index Price Stream
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Index-Price-Stream
    """
    stream_name = "{}@indexPrice".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mark_price(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    Mark price update stream
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Mark-Price-Stream
    """
    stream_name = "{}@markPrice".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def kline(
    self,
    symbol: str,
    interval: Literal[
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "8h",
        "12h",
        "1d",
        "3d",
        "1w",
        "1M",
    ],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    The Kline/Candlestick Stream push updates to the current klines/candlestick every 250 milliseconds (if existing).
    Kline/Candlestick chart intervals: m -> minutes; h -> hours; d -> days; w -> weeks; M -> months
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Kline-Candlestick-Streams
    """
    stream_name = "{}@kline_{}".format(symbol.lower(), interval)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def continuous_kline(
    self,
    pair: str,
    contract_type: Literal["perpetual", "current_quarter", "next_quarter"],
    interval: Literal[
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "8h",
        "12h",
        "1d",
        "3d",
        "1w",
        "1M",
    ],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    Kline update every second
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Continuous-Contract-Kline-Candlestick-Streams
    """
    stream_name = "{}_{}@continuousKline_{}".format(
        pair.lower(), contract_type, interval
    )
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def pair_mark_price(
    self, pair: str, subscription: bool = True, _id: int | None = None
):
    """
    Mark Price of All Symbols of a Pair
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Mark-Price-of-All-Symbols-of-a-Pair
    """
    stream_name = "{}@markPrice".format(pair.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def index_price_kline(
    self,
    pair: str,
    interval: Literal[
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "8h",
        "12h",
        "1d",
        "3d",
        "1w",
        "1M",
    ],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    Index Kline/Candlestick Streams
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Index-Kline-Candlestick-Streams
    """
    stream_name = "{}@indexPriceKline_{}".format(pair.lower(), interval)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mark_price_kline(
    self,
    symbol: str,
    interval: Literal[
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "8h",
        "12h",
        "1d",
        "3d",
        "1w",
        "1M",
    ],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    Mark Price Kline/Candlestick Streams
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Mark-Price-Kline-Candlestick-Streams
    """
    stream_name = "{}@markPriceKline_{}".format(symbol.lower(), interval)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mini_ticker(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    24hr rolling window mini-ticker statistics for a single symbol. These are NOT the statistics of the UTC day, but a 24hr rolling window from requestTime to 24hrs before.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Individual-Symbol-Mini-Ticker-Stream
    """
    stream_name = "{}@miniTicker".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mini_ticker_arr(self, subscription: bool = True, _id: int | None = None):
    """
    24hr rolling window mini-ticker statistics for all symbols. These are NOT the statistics of the UTC day, but a 24hr rolling window from requestTime to 24hrs before. Note that only tickers that have changed will be present in the array.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/All-Market-Mini-Tickers-Stream
    """
    stream_name = "!miniTicker@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def ticker(self, symbol: str, subscription: bool = True, _id: int | None = None):
    """
    24hr rolling window ticker statistics for a single symbol. These are NOT the statistics of the UTC day, but a 24hr rolling window from requestTime to 24hrs before.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Individual-Symbol-Ticker-Streams
    """
    stream_name = "{}@ticker".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def ticker_arr(self, subscription: bool = True, _id: int | None = None):
    """
    24hr rolling window ticker statistics for all symbols. These are NOT the statistics of the UTC day, but a 24hr rolling window from requestTime to 24hrs before. Note that only tickers that have changed will be present in the array.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/All-Market-Tickers-Streams
    """
    stream_name = "!ticker@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def book_ticker(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    Pushes any update to the best bid or ask's price or quantity in real-time for a specified symbol.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Individual-Symbol-Book-Ticker-Streams
    """
    stream_name = "{}@bookTicker".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def book_ticker_arr(self, subscription: bool = True, _id: int | None = None):
    """
    Pushes any update to the best bid or ask's price or quantity in real-time for all symbols.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/All-Book-Tickers-Stream
    """
    stream_name = "!bookTicker"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def force_orders(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    The Liquidation Order Snapshot Streams push force liquidation order information for specific symbol.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Liquidation-Order-Streams
    """
    stream_name = "{}@forceOrder".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def force_orders_arr(self, subscription: bool = True, _id: int | None = None):
    """
    The All Liquidation Order Snapshot Streams push force liquidation order information for all symbols in the market. For each symbol，only the latest one liquidation order within 1000ms will be pushed as the snapshot. If no liquidation happens in the interval of 1000ms, no stream will be pushed.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/All-Market-Liquidation-Order-Streams
    """
    stream_name = "!forceOrder@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def contract_info(self, subscription: bool = True, _id: int | None = None):
    """
    ContractInfo stream pushes when contract info updates(listing/settlement/contract bracket update). bks field only shows up when bracket gets updated.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Contract-Info-Stream
    """
    stream_name = "!contractInfo"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def depth_levels(
    self,
    symbol: str,
    levels: Literal[5, 10, 20],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    Top <levels> bids and asks, Valid <levels> are 5, 10, or 20.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Partial-Book-Depth-Streams
    """
    stream_name = "{}@depth{}".format(symbol.lower(), levels)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def diff_depth(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    Bids and asks, pushed every 250 milliseconds, 500 milliseconds, or 100 milliseconds
    https://developers.binance.com/docs/derivatives/coin-margined-futures/websocket-market-streams/Diff-Book-Depth-Streams
    """
    stream_name = "{}@depth".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)



from typing import Literal


async def get_dapi_v1_balance(self, **kwargs):
    """
    Query account balance info
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Futures-Account-Balance
    """
    endpoint = "/dapi/v1/balance"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_account(self, **kwargs):
    """
    Get current account information. User in single-asset/ multi-assets mode will see different value, see comments in response section for detail.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Account-Information
    """
    endpoint = "/dapi/v1/account"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_commission_rate(self, symbol: str, **kwargs):
    """
    Query user commission rate
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/User-Commission-Rate
    """
    endpoint = "/dapi/v1/commissionRate"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v2_leverage_bracket(self, **kwargs):
    """
    Get the symbol's notional bracket list.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Notional-Bracket-for-Symbol
    """
    endpoint = "/dapi/v2/leverageBracket"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_leverage_bracket(self, **kwargs):
    """
    Not recommended to continue using this v1 endpoint
    Get the pair's default notional bracket list, may return ambiguous values when there have been multiple different symbol brackets under the pair, suggest using the following GET /dapi/v2/leverageBracket query instead to get the specific symbol notional bracket list.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Notional-Bracket-for-Pair
    """
    endpoint = "/dapi/v1/leverageBracket"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_position_side_dual(self, **kwargs):
    """
    Get user's position mode (Hedge Mode or One-way Mode ) on EVERY symbol
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Get-Current-Position-Mode
    """
    endpoint = "/dapi/v1/positionSide/dual"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_income(self, **kwargs):
    """
    Query income history
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Get-Income-History
    """
    endpoint = "/dapi/v1/income"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_income_asyn(self, start_time: int, end_time: int, **kwargs):
    """
    Get download id for futures transaction history
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Get-Download-Id-For-Futures-Transaction-History
    """
    endpoint = "/dapi/v1/income/asyn"
    payload = {"startTime": start_time, "endTime": end_time, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_income_asyn_id(self, download_id: str, **kwargs):
    """
    Get futures transaction history download link by Id
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Get-Futures-Transaction-History-Download-Link-by-Id
    """
    endpoint = "/dapi/v1/income/asyn/id"
    payload = {"downloadId": download_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_order_asyn(self, start_time: int, end_time: int, **kwargs):
    """
    Get Download Id For Futures Order History
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Get-Download-Id-For-Futures-Order-History
    """
    endpoint = "/dapi/v1/order/asyn"
    payload = {"startTime": start_time, "endTime": end_time, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_order_asyn_id(self, download_id: str, **kwargs):
    """
    Get futures order history download link by Id
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Get-Futures-Order-History-Download-Link-by-Id
    """
    endpoint = "/dapi/v1/order/asyn/id"
    payload = {"downloadId": download_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_trade_asyn(self, start_time: int, end_time: int, **kwargs):
    """
    Get download id for futures trade history
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Get-Download-Id-For-Futures-Trade-History
    """
    endpoint = "/dapi/v1/trade/asyn"
    payload = {"startTime": start_time, "endTime": end_time, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_trade_asyn_id(self, download_id: str, **kwargs):
    """
    Get futures trade download link by Id
    https://developers.binance.com/docs/derivatives/coin-margined-futures/account/Get-Futures-Trade-Download-Link-by-Id
    """
    endpoint = "/dapi/v1/trade/asyn/id"
    payload = {"downloadId": download_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_pm_account_info(self, asset: str, **kwargs):
    """
    Get Classic Portfolio Margin current account information.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/portfolio-margin-endpoints
    """
    endpoint = "/dapi/v1/pmAccountInfo"
    payload = {"asset": asset, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw



async def get_sapi_v1_simpleEarn_flexible_list(
    self, asset: str | None = None, current: int | None = None, size: int | None = None
):
    """
    GET /sapi/v1/simple-earn/flexible/list

    Request Weight(IP)
    150

    Request Parameters
    Name	Type	Mandatory	Description
    asset	STRING	NO
    current	LONG	NO	Currently querying page. Start from 1. Default:1
    size	LONG	NO	Default:10, Max:100
    recvWindow	LONG	NO
    timestamp	LONG	YES
    """
    endpoint = "/sapi/v1/simple-earn/flexible/list"
    payload = {
        "asset": asset,
        "current": current,
        "size": size,
    }
    payload = {k: v for k, v in payload.items() if v is not None}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_sapi_v1_simpleEarn_locked_list(
    self, asset: str | None = None, current: int | None = None, size: int | None = None
):
    """
    GET /sapi/v1/simple-earn/locked/list

    Request Weight(IP)
    150

    Request Parameters
    Name	Type	Mandatory	Description
    asset	STRING	NO
    current	LONG	NO	Currently querying page. Start from 1. Default:1
    size	LONG	NO	Default:10, Max:100
    recvWindow	LONG	NO
    timestamp	LONG	YES
    """

    endpoint = "/sapi/v1/simple-earn/locked/list"
    payload = {
        "asset": asset,
        "current": current,
        "size": size,
    }
    payload = {k: v for k, v in payload.items() if v is not None}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw



from enum import Enum


class BybitUrl(Enum):
    MAINNET_1 = "MAINNET_1"
    MAINNET_2 = "MAINNET_2"
    TESTNET = "TESTNET"
    NETHERLAND = "NETHERLAND"
    HONGKONG = "HONGKONG"
    TURKEY = "TURKEY"
    HAZAKHSTAN = "HAZAKHSTAN"

    @property
    def base_url(self):
        return BASE_URLS[self]

    @property
    def is_testnet(self):
        return self == BybitUrl.TESTNET


BASE_URLS = {
    BybitUrl.MAINNET_1: "https://api.bybit.com",
    BybitUrl.MAINNET_2: "https://api.bytick.com",
    BybitUrl.TESTNET: "https://api-testnet.bybit.com",
    BybitUrl.NETHERLAND: "https://api.bybit.nl",
    BybitUrl.HONGKONG: "https://api.byhkbit.com",
    BybitUrl.TURKEY: "https://api.bybit-tr.com",
    BybitUrl.HAZAKHSTAN: "https://api.bybit.kz",
}


class BybitAccountType(Enum):
    SPOT = "SPOT"
    LINEAR = "LINEAR"
    INVERSE = "INVERSE"
    OPTION = "OPTION"
    SPOT_TESTNET = "SPOT_TESTNET"
    LINEAR_TESTNET = "LINEAR_TESTNET"
    INVERSE_TESTNET = "INVERSE_TESTNET"
    OPTION_TESTNET = "OPTION_TESTNET"
    UNIFIED = "UNIFIED"
    UNIFIED_TESTNET = "UNIFIED_TESTNET"

    @property
    def exchange_id(self):
        return "bybit"

    @property
    def is_testnet(self):
        return self in {
            self.SPOT_TESTNET,
            self.LINEAR_TESTNET,
            self.INVERSE_TESTNET,
            self.OPTION_TESTNET,
            self.UNIFIED_TESTNET,
        }

    @property
    def ws_public_url(self):
        return WS_PUBLIC_URL[self]

    @property
    def ws_private_url(self):
        if self.is_testnet:
            return "wss://stream-testnet.bybit.com/v5/private"
        return "wss://stream.bybit.com/v5/private"

    @property
    def is_spot(self):
        return self in {self.SPOT, self.SPOT_TESTNET}

    @property
    def is_linear(self):
        return self in {self.LINEAR, self.LINEAR_TESTNET}

    @property
    def is_inverse(self):
        return self in {self.INVERSE, self.INVERSE_TESTNET}


WS_PUBLIC_URL = {
    BybitAccountType.SPOT: "wss://stream.bybit.com/v5/public/spot",
    BybitAccountType.LINEAR: "wss://stream.bybit.com/v5/public/linear",
    BybitAccountType.INVERSE: "wss://stream.bybit.com/v5/public/inverse",
    BybitAccountType.OPTION: "wss://stream.bybit.com/v5/public/option",
    BybitAccountType.SPOT_TESTNET: "wss://stream-testnet.bybit.com/v5/public/spot",
    BybitAccountType.LINEAR_TESTNET: "wss://stream-testnet.bybit.com/v5/public/linear",
    BybitAccountType.INVERSE_TESTNET: "wss://stream-testnet.bybit.com/v5/public/inverse",
    BybitAccountType.OPTION_TESTNET: "wss://stream-testnet.bybit.com/v5/public/option",
}



from typing import Literal

from nexusbridge.binance.constants import Interval, ContractType


async def get_dapi_v1_ping(
    self,
):
    """
    Test connectivity to the Rest API.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data
    """
    endpoint = "/dapi/v1/ping"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_dapi_v1_time(
    self,
):
    """
    Test connectivity to the Rest API and get the current server time.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Check-Server-time
    """
    endpoint = "/dapi/v1/time"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_dapi_v1_exchange_info(
    self,
):
    """
    Current exchange trading rules and symbol information
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Exchange-Information
    """
    endpoint = "/dapi/v1/exchangeInfo"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_dapi_v1_depth(self, symbol: str, **kwargs):
    """
    Query orderbook on specific symbol
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Order-Book
    """
    endpoint = "/dapi/v1/depth"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_trades(self, symbol: str, **kwargs):
    """
    Get recent market trades
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Recent-Trades-List
    """
    endpoint = "/dapi/v1/trades"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_historical_trades(self, symbol: str, **kwargs):
    """
    Get older market historical trades.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Old-Trades-Lookup
    """
    endpoint = "/dapi/v1/historicalTrades"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_agg_trades(self, symbol: str, **kwargs):
    """
    Get compressed, aggregate trades. Market trades that fill in 100ms with the same price and the same taking side will have the quantity aggregated.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Compressed-Aggregate-Trades-List
    """
    endpoint = "/dapi/v1/aggTrades"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_premium_index(self, **kwargs):
    """
    Query index price and mark price
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Index-Price-and-Mark-Price
    """
    endpoint = "/dapi/v1/premiumIndex"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_funding_rate(self, symbol: str, **kwargs):
    """
    Get Funding Rate History of Perpetual Futures
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Get-Funding-Rate-History-of-Perpetual-Futures
    """
    endpoint = "/dapi/v1/fundingRate"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_funding_info(
    self,
):
    """
    Query funding rate info for symbols that had FundingRateCap/ FundingRateFloor / fundingIntervalHours adjustment
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Get-Funding-Info
    """
    endpoint = "/dapi/v1/fundingInfo"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw


async def get_dapi_v1_klines(self, symbol: str, interval: Interval, **kwargs):
    """
    Kline/candlestick bars for a symbol.
    Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/kline-candlestick-data
    """
    endpoint = "/dapi/v1/klines"
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_continuous_klines(
    self, pair: str, interval: Interval, contract_type: ContractType, **kwargs
):
    """
    Kline/candlestick bars for a specific contract type. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Continuous-Contract-Kline-Candlestick-Data
    """
    endpoint = "/dapi/v1/continuousKlines"
    payload = {
        "pair": pair,
        "contractType": contract_type.value,
        "interval": interval.value,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_index_price_klines(self, pair: str, interval: Interval, **kwargs):
    """
    Kline/candlestick bars for the index price of a pair. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Index-Price-Kline-Candlestick-Data
    """
    endpoint = "/dapi/v1/indexPriceKlines"
    payload = {"pair": pair, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_mark_price_klines(
    self, symbol: str, interval: Interval, **kwargs
):
    """
    Kline/candlestick bars for the mark price of a symbol. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Mark-Price-Kline-Candlestick-Data
    """
    endpoint = "/dapi/v1/markPriceKlines"
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_premium_index_klines(
    self, symbol: str, interval: Interval, **kwargs
):
    """
    Premium index kline bars of a symbol. Klines are uniquely identified by their open time.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Premium-Index-Kline-Data
    """
    endpoint = "/dapi/v1/premiumIndexKlines"
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_ticker_24hr(self, **kwargs):
    """
    24 hour rolling window price change statistics.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/24hr-Ticker-Price-Change-Statistics
    """
    endpoint = "/dapi/v1/ticker/24hr"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_ticker_price(self, **kwargs):
    """
    Latest price for a symbol or symbols.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Symbol-Price-Ticker
    """
    endpoint = "/dapi/v1/ticker/price"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_ticker_book_ticker(self, **kwargs):
    """
    Best price/qty on the order book for a symbol or symbols.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Symbol-Order-Book-Ticker
    """
    endpoint = "/dapi/v1/ticker/bookTicker"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_open_interest(
    self,
    symbol: str,
):
    """
    Get present open interest of a specific symbol.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Open-Interest
    """
    endpoint = "/dapi/v1/openInterest"
    payload = {
        "symbol": symbol,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_futures_data_open_interest_hist(
    self,
    pair: str,
    contract_type: ContractType,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    Query open interest stats
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Open-Interest-Statistics
    """
    endpoint = "/futures/data/openInterestHist"
    payload = {
        "pair": pair,
        "contractType": contract_type.value,
        "period": period,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_futures_data_top_long_short_position_ratio(
    self,
    pair: str,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    The proportion of net long and net short positions to total open positions of the top 20% users with the highest margin balance. Long Position % = Long positions of top traders / Total open positions of top traders Short Position % = Short positions of top traders / Total open positions of top traders Long/Short Ratio (Positions) = Long Position % / Short Position %
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Top-Trader-Long-Short-Ratio
    """
    endpoint = "/futures/data/topLongShortPositionRatio"
    payload = {"pair": pair, "period": period, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_futures_data_top_long_short_account_ratio(
    self,
    pair: str,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    The proportion of net long and net short accounts to total accounts of the top 20% users with the highest margin balance. Each account is counted once only. Long Account % = Accounts of top traders with net long positions / Total accounts of top traders with open positions Short Account % = Accounts of top traders with net short positions / Total accounts of top traders with open positions Long/Short Ratio (Accounts) = Long Account % / Short Account %
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Top-Long-Short-Account-Ratio
    """
    endpoint = "/futures/data/topLongShortAccountRatio"
    payload = {"pair": pair, "period": period, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_futures_data_global_long_short_account_ratio(
    self,
    pair: str,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    Query symbol Long/Short Ratio
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Long-Short-Ratio
    """
    endpoint = "/futures/data/globalLongShortAccountRatio"
    payload = {"pair": pair, "period": period, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_futures_data_taker_buy_sell_vol(
    self,
    pair: str,
    contract_type: ContractType,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    Taker Buy Volume: the total volume of buy orders filled by takers within the period. Taker Sell Volume: the total volume of sell orders filled by takers within the period.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Taker-Buy-Sell-Volume
    """
    endpoint = "/futures/data/takerBuySellVol"
    payload = {
        "pair": pair,
        "period": period,
        "contractType": contract_type.value,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_futures_data_basis(
    self,
    pair: str,
    contract_type: ContractType,
    period: Literal["5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d"],
    **kwargs,
):
    """
    Query Basis
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Basis
    """
    endpoint = "/futures/data/basis"
    payload = {
        "pair": pair,
        "period": period,
        "contractType": contract_type.value,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def get_dapi_v1_constituents(
    self,
    symbol: str,
):
    """
    Query index price constituents
    https://developers.binance.com/docs/derivatives/coin-margined-futures/market-data/Index-Constituents
    """
    endpoint = "/dapi/v1/constituents"
    payload = {
        "symbol": symbol,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw
